from typing import Iterable, List, Optional

from pyzeebe.errors import NoVariableNameGivenError
from pyzeebe.function_tools import async_tools
from pyzeebe.task.exception_handler import ExceptionHandler
from pyzeebe.task.types import AsyncTaskDecorator, TaskDecorator


class TaskConfig:
    before: List[AsyncTaskDecorator]
    after: List[AsyncTaskDecorator]

    def __init__(
        self,
        type: str,
        exception_handler: Optional[ExceptionHandler],
        timeout_ms: int,
        max_jobs_to_activate: int,
        max_running_jobs: int,
        variables_to_fetch: Optional[Iterable[str]],
        single_value: bool,
        variable_name: str,
        before: List[TaskDecorator],
        after: List[TaskDecorator],
    ) -> None:
        if single_value and not variable_name:
            raise NoVariableNameGivenError(type)

        self.type = type
        self.exception_handler = exception_handler
        self.timeout_ms = timeout_ms
        self.max_jobs_to_activate = max_jobs_to_activate
        self.max_running_jobs = max_running_jobs
        self.variables_to_fetch = variables_to_fetch
        self.single_value = single_value
        self.variable_name = variable_name
        self.before = async_tools.asyncify_all_functions(before)
        self.after = async_tools.asyncify_all_functions(after)
        self.job_parameter_name: Optional[str] = None

    def __repr__(self) -> str:
        return (
            f"TaskConfig(type={self.type}, exception_handler={self.exception_handler}, "
            f"timeout_ms={self.timeout_ms}, max_jobs_to_activate={self.max_jobs_to_activate}, "
            f"max_running_jobs={self.max_running_jobs}, variables_to_fetch={self.variables_to_fetch},"
            f"single_value={self.single_value}, variable_name={self.variable_name},"
            f"before={self.before}, after={self.after})"
        )



import grpc

from pyzeebe.errors.pyzeebe_errors import PyZeebeError


class ZeebeBackPressureError(PyZeebeError):
    pass


class ZeebeGatewayUnavailableError(PyZeebeError):
    pass


class ZeebeInternalError(PyZeebeError):
    pass


class UnknownGrpcStatusCodeError(PyZeebeError):
    def __init__(self, grpc_error: grpc.aio.AioRpcError):
        super().__init__()
        self.grpc_error = grpc_error



import abc

from pyzeebe.credentials.typing import AuthMetadata, CallContext


class CredentialsABC(abc.ABC):
    """A specification for credentials manager. Passed to :py:class:`AuthMetadataPlugin`."""

    @abc.abstractmethod
    def get_auth_metadata(self, context: CallContext) -> AuthMetadata:
        """
        Args:
            context (grpc.AuthMetadataContext): Provides information to call credentials metadata plugins.

        Returns:
            Tuple[Tuple[str, Union[str, bytes]], ...]: The `metadata` used to construct the :py:class:`grpc.CallCredentials`.
        """
        raise NotImplementedError



import logging
from typing import NoReturn

import grpc
from zeebe_grpc.gateway_pb2_grpc import GatewayStub

from pyzeebe.errors import (
    UnknownGrpcStatusCodeError,
    ZeebeBackPressureError,
    ZeebeGatewayUnavailableError,
    ZeebeInternalError,
)
from pyzeebe.errors.pyzeebe_errors import PyZeebeError
from pyzeebe.grpc_internals.grpc_utils import is_error_status

logger = logging.getLogger(__name__)


class ZeebeAdapterBase:
    def __init__(self, grpc_channel: grpc.aio.Channel, max_connection_retries: int = -1):
        self._channel = grpc_channel
        self._gateway_stub = GatewayStub(grpc_channel)
        self.connected = True
        self.retrying_connection = False
        self._max_connection_retries = max_connection_retries
        self._current_connection_retries = 0

    def _should_retry(self) -> bool:
        return self._max_connection_retries == -1 or self._current_connection_retries < self._max_connection_retries

    async def _handle_grpc_error(self, grpc_error: grpc.aio.AioRpcError) -> NoReturn:
        try:
            pyzeebe_error = _create_pyzeebe_error_from_grpc_error(grpc_error)
            raise pyzeebe_error
        except (ZeebeGatewayUnavailableError, ZeebeInternalError):
            self._current_connection_retries += 1
            if not self._should_retry():
                await self._close()
            raise

    async def _close(self) -> None:
        try:
            await self._channel.close()
        except Exception as exception:
            logger.exception("Failed to close channel, %s exception was raised", type(exception).__name__)


def _create_pyzeebe_error_from_grpc_error(grpc_error: grpc.aio.AioRpcError) -> PyZeebeError:
    if is_error_status(grpc_error, grpc.StatusCode.RESOURCE_EXHAUSTED):
        return ZeebeBackPressureError()
    if is_error_status(grpc_error, grpc.StatusCode.UNAVAILABLE, grpc.StatusCode.CANCELLED):
        return ZeebeGatewayUnavailableError()
    if is_error_status(grpc_error, grpc.StatusCode.INTERNAL):
        return ZeebeInternalError()
    return UnknownGrpcStatusCodeError(grpc_error)



from enum import Enum


class JobStatus(Enum):
    Running = "Running"
    Completed = "Completed"
    RunningAfterDecorators = "RunningAfterDecorators"
    Failed = "Failed"
    ErrorThrown = "ErrorThrown"



import json
from typing import Optional

import grpc
from zeebe_grpc.gateway_pb2 import PublishMessageRequest

from pyzeebe.errors import MessageAlreadyExistsError
from pyzeebe.grpc_internals.grpc_utils import is_error_status
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase
from pyzeebe.types import Variables

from .types import PublishMessageResponse


class ZeebeMessageAdapter(ZeebeAdapterBase):
    async def publish_message(
        self,
        name: str,
        correlation_key: str,
        time_to_live_in_milliseconds: int,
        variables: Variables,
        message_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
    ) -> PublishMessageResponse:
        try:
            response = await self._gateway_stub.PublishMessage(
                PublishMessageRequest(
                    name=name,
                    correlationKey=correlation_key,
                    messageId=message_id,
                    timeToLive=time_to_live_in_milliseconds,
                    variables=json.dumps(variables),
                    tenantId=tenant_id,
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.ALREADY_EXISTS):
                raise MessageAlreadyExistsError() from grpc_error
            await self._handle_grpc_error(grpc_error)

        return PublishMessageResponse(key=response.key, tenant_id=response.tenantId)



import functools
from typing import Any, Dict, TypeVar

from typing_extensions import ParamSpec

from pyzeebe.function_tools import AsyncFunction, DictFunction

P = ParamSpec("P")
R = TypeVar("R")


def convert_to_dict_function(single_value_function: AsyncFunction[P, R], variable_name: str) -> DictFunction[P]:
    @functools.wraps(single_value_function)
    async def inner_fn(*args: P.args, **kwargs: P.kwargs) -> Dict[str, Any]:
        return {variable_name: await single_value_function(*args, **kwargs)}

    return inner_fn



import json
import logging
from typing import AsyncGenerator, Iterable, Optional

import grpc
from zeebe_grpc.gateway_pb2 import (
    ActivatedJob,
    ActivateJobsRequest,
    CompleteJobRequest,
    FailJobRequest,
    ThrowErrorRequest,
)

from pyzeebe.errors import (
    ActivateJobsRequestInvalidError,
    JobAlreadyDeactivatedError,
    JobNotFoundError,
)
from pyzeebe.grpc_internals.grpc_utils import is_error_status
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase
from pyzeebe.job.job import Job
from pyzeebe.types import Variables

from .types import CompleteJobResponse, FailJobResponse, ThrowErrorResponse

logger = logging.getLogger(__name__)


class ZeebeJobAdapter(ZeebeAdapterBase):
    async def activate_jobs(
        self,
        task_type: str,
        worker: str,
        timeout: int,
        max_jobs_to_activate: int,
        variables_to_fetch: Iterable[str],
        request_timeout: int,
        tenant_ids: Optional[Iterable[str]] = None,
    ) -> AsyncGenerator[Job, None]:
        try:
            async for response in self._gateway_stub.ActivateJobs(
                ActivateJobsRequest(
                    type=task_type,
                    worker=worker,
                    timeout=timeout,
                    maxJobsToActivate=max_jobs_to_activate,
                    fetchVariable=variables_to_fetch,
                    requestTimeout=request_timeout,
                    tenantIds=tenant_ids,
                )
            ):
                for raw_job in response.jobs:
                    job = self._create_job_from_raw_job(raw_job)
                    logger.debug("Got job: %s from zeebe", job)
                    yield job
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.INVALID_ARGUMENT):
                raise ActivateJobsRequestInvalidError(task_type, worker, timeout, max_jobs_to_activate) from grpc_error
            await self._handle_grpc_error(grpc_error)

    def _create_job_from_raw_job(self, response: ActivatedJob) -> Job:
        return Job(
            key=response.key,
            type=response.type,
            process_instance_key=response.processInstanceKey,
            bpmn_process_id=response.bpmnProcessId,
            process_definition_version=response.processDefinitionVersion,
            process_definition_key=response.processDefinitionKey,
            element_id=response.elementId,
            element_instance_key=response.elementInstanceKey,
            custom_headers=json.loads(response.customHeaders),
            worker=response.worker,
            retries=response.retries,
            deadline=response.deadline,
            variables=json.loads(response.variables),
            tenant_id=response.tenantId,
            zeebe_adapter=self,  # type: ignore[arg-type]
        )

    async def complete_job(self, job_key: int, variables: Variables) -> CompleteJobResponse:
        try:
            await self._gateway_stub.CompleteJob(CompleteJobRequest(jobKey=job_key, variables=json.dumps(variables)))
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key) from grpc_error
            elif is_error_status(grpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key) from grpc_error
            await self._handle_grpc_error(grpc_error)

        return CompleteJobResponse()

    async def fail_job(
        self, job_key: int, retries: int, message: str, retry_back_off_ms: int, variables: Variables
    ) -> FailJobResponse:
        try:
            await self._gateway_stub.FailJob(
                FailJobRequest(
                    jobKey=job_key,
                    retries=retries,
                    errorMessage=message,
                    retryBackOff=retry_back_off_ms,
                    variables=json.dumps(variables),
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key) from grpc_error
            elif is_error_status(grpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key) from grpc_error
            await self._handle_grpc_error(grpc_error)

        return FailJobResponse()

    async def throw_error(
        self, job_key: int, message: str, variables: Variables, error_code: str = ""
    ) -> ThrowErrorResponse:
        try:
            await self._gateway_stub.ThrowError(
                ThrowErrorRequest(
                    jobKey=job_key,
                    errorMessage=message,
                    errorCode=error_code,
                    variables=json.dumps(variables),
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.NOT_FOUND):
                raise JobNotFoundError(job_key=job_key) from grpc_error
            elif is_error_status(grpc_error, grpc.StatusCode.FAILED_PRECONDITION):
                raise JobAlreadyDeactivatedError(job_key=job_key) from grpc_error
            await self._handle_grpc_error(grpc_error)

        return ThrowErrorResponse()



import json
import os
from typing import Any, Callable, Dict, Iterable, List, NoReturn, Optional, Union

import aiofiles
import grpc
from typing_extensions import deprecated
from zeebe_grpc.gateway_pb2 import (
    CancelProcessInstanceRequest,
    CreateProcessInstanceRequest,
    CreateProcessInstanceWithResultRequest,
    DecisionMetadata,
    DecisionRequirementsMetadata,
    DeployProcessRequest,
    DeployResourceRequest,
    FormMetadata,
    ProcessMetadata,
    ProcessRequestObject,
    Resource,
)

from pyzeebe.errors import (
    InvalidJSONError,
    ProcessDefinitionHasNoStartEventError,
    ProcessDefinitionNotFoundError,
    ProcessInstanceNotFoundError,
    ProcessInvalidError,
    ProcessTimeoutError,
)
from pyzeebe.grpc_internals.grpc_utils import is_error_status
from pyzeebe.grpc_internals.zeebe_adapter_base import ZeebeAdapterBase
from pyzeebe.types import Variables

from .types import (
    CancelProcessInstanceResponse,
    CreateProcessInstanceResponse,
    CreateProcessInstanceWithResultResponse,
    DeployProcessResponse,
    DeployResourceResponse,
)


class ZeebeProcessAdapter(ZeebeAdapterBase):
    async def create_process_instance(
        self,
        bpmn_process_id: str,
        version: int,
        variables: Variables,
        tenant_id: Optional[str] = None,
    ) -> CreateProcessInstanceResponse:
        try:
            response = await self._gateway_stub.CreateProcessInstance(
                CreateProcessInstanceRequest(
                    bpmnProcessId=bpmn_process_id,
                    version=version,
                    variables=json.dumps(variables),
                    tenantId=tenant_id,
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            await self._create_process_errors(grpc_error, bpmn_process_id, version, variables)

        return CreateProcessInstanceResponse(
            process_definition_key=response.processDefinitionKey,
            bpmn_process_id=response.bpmnProcessId,
            version=response.version,
            process_instance_key=response.processInstanceKey,
            tenant_id=response.tenantId,
        )

    async def create_process_instance_with_result(
        self,
        bpmn_process_id: str,
        version: int,
        variables: Variables,
        timeout: int,
        variables_to_fetch: Iterable[str],
        tenant_id: Optional[str] = None,
    ) -> CreateProcessInstanceWithResultResponse:
        try:
            response = await self._gateway_stub.CreateProcessInstanceWithResult(
                CreateProcessInstanceWithResultRequest(
                    request=CreateProcessInstanceRequest(
                        bpmnProcessId=bpmn_process_id,
                        version=version,
                        variables=json.dumps(variables),
                        tenantId=tenant_id,
                    ),
                    requestTimeout=timeout,
                    fetchVariables=variables_to_fetch,
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            await self._create_process_errors(grpc_error, bpmn_process_id, version, variables)

        return CreateProcessInstanceWithResultResponse(
            process_definition_key=response.processDefinitionKey,
            bpmn_process_id=response.bpmnProcessId,
            version=response.version,
            process_instance_key=response.processInstanceKey,
            variables=json.loads(response.variables),
            tenant_id=response.tenantId,
        )

    async def _create_process_errors(
        self, grpc_error: grpc.aio.AioRpcError, bpmn_process_id: str, version: int, variables: Dict[str, Any]
    ) -> NoReturn:
        if is_error_status(grpc_error, grpc.StatusCode.NOT_FOUND):
            raise ProcessDefinitionNotFoundError(bpmn_process_id=bpmn_process_id, version=version) from grpc_error
        elif is_error_status(grpc_error, grpc.StatusCode.INVALID_ARGUMENT):
            raise InvalidJSONError(
                f"Cannot start process: {bpmn_process_id} with version {version}. Variables: {variables}"
            ) from grpc_error
        elif is_error_status(grpc_error, grpc.StatusCode.FAILED_PRECONDITION):
            raise ProcessDefinitionHasNoStartEventError(bpmn_process_id=bpmn_process_id) from grpc_error
        elif is_error_status(grpc_error, grpc.StatusCode.DEADLINE_EXCEEDED):
            raise ProcessTimeoutError(bpmn_process_id) from grpc_error
        await self._handle_grpc_error(grpc_error)

    async def cancel_process_instance(self, process_instance_key: int) -> CancelProcessInstanceResponse:
        try:
            await self._gateway_stub.CancelProcessInstance(
                CancelProcessInstanceRequest(processInstanceKey=process_instance_key)
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.NOT_FOUND):
                raise ProcessInstanceNotFoundError(process_instance_key=process_instance_key) from grpc_error
            await self._handle_grpc_error(grpc_error)

        return CancelProcessInstanceResponse()

    @deprecated("Deprecated since Zeebe 8.0. Use deploy_resource instead")
    async def deploy_process(self, *process_file_path: str) -> DeployProcessResponse:
        try:
            response = await self._gateway_stub.DeployProcess(
                DeployProcessRequest(
                    processes=[await result for result in map(_create_process_request, process_file_path)]
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.INVALID_ARGUMENT):
                raise ProcessInvalidError() from grpc_error
            await self._handle_grpc_error(grpc_error)

        return DeployProcessResponse(
            key=response.key,
            processes=[
                DeployProcessResponse.ProcessMetadata(
                    bpmn_process_id=process.bpmnProcessId,
                    version=process.version,
                    process_definition_key=process.processDefinitionKey,
                    resource_name=process.resourceName,
                )
                for process in response.processes
            ],
        )

    async def deploy_resource(
        self, *resource_file_path: str, tenant_id: Optional[str] = None
    ) -> DeployResourceResponse:
        try:
            response = await self._gateway_stub.DeployResource(
                DeployResourceRequest(
                    resources=[await result for result in map(_create_resource_request, resource_file_path)],
                    tenantId=tenant_id,
                )
            )
        except grpc.aio.AioRpcError as grpc_error:
            if is_error_status(grpc_error, grpc.StatusCode.INVALID_ARGUMENT):
                raise ProcessInvalidError() from grpc_error
            await self._handle_grpc_error(grpc_error)

        deployments: List[
            Union[
                DeployResourceResponse.ProcessMetadata,
                DeployResourceResponse.DecisionMetadata,
                DeployResourceResponse.DecisionRequirementsMetadata,
                DeployResourceResponse.FormMetadata,
            ]
        ] = []
        for deployment in response.deployments:
            metadata_field = deployment.WhichOneof("Metadata")
            metadata = getattr(deployment, metadata_field)
            deployments.append(_METADATA_PARSERS[metadata_field](metadata))

        return DeployResourceResponse(
            key=response.key,
            deployments=deployments,
            tenant_id=response.tenantId,
        )

    @staticmethod
    def _create_process_from_raw_process(response: ProcessMetadata) -> DeployResourceResponse.ProcessMetadata:
        return DeployResourceResponse.ProcessMetadata(
            bpmn_process_id=response.bpmnProcessId,
            version=response.version,
            process_definition_key=response.processDefinitionKey,
            resource_name=response.resourceName,
            tenant_id=response.tenantId,
        )

    @staticmethod
    def _create_decision_from_raw_decision(response: DecisionMetadata) -> DeployResourceResponse.DecisionMetadata:
        return DeployResourceResponse.DecisionMetadata(
            dmn_decision_id=response.dmnDecisionId,
            dmn_decision_name=response.dmnDecisionName,
            version=response.version,
            decision_key=response.decisionKey,
            dmn_decision_requirements_id=response.dmnDecisionRequirementsId,
            decision_requirements_key=response.decisionRequirementsKey,
            tenant_id=response.tenantId,
        )

    @staticmethod
    def _create_decision_requirements_from_raw_decision_requirements(
        response: DecisionRequirementsMetadata,
    ) -> DeployResourceResponse.DecisionRequirementsMetadata:
        return DeployResourceResponse.DecisionRequirementsMetadata(
            dmn_decision_requirements_id=response.dmnDecisionRequirementsId,
            dmn_decision_requirements_name=response.dmnDecisionRequirementsName,
            version=response.version,
            decision_requirements_key=response.decisionRequirementsKey,
            resource_name=response.resourceName,
            tenant_id=response.tenantId,
        )

    @staticmethod
    def _create_form_from_raw_form(response: FormMetadata) -> DeployResourceResponse.FormMetadata:
        return DeployResourceResponse.FormMetadata(
            form_id=response.formId,
            version=response.version,
            form_key=response.formKey,
            resource_name=response.resourceName,
            tenant_id=response.tenantId,
        )


_METADATA_PARSERS: Dict[
    str,
    Callable[
        [Union[ProcessMetadata, DecisionMetadata, DecisionRequirementsMetadata, FormMetadata]],
        Union[
            DeployResourceResponse.ProcessMetadata,
            DeployResourceResponse.DecisionMetadata,
            DeployResourceResponse.DecisionRequirementsMetadata,
            DeployResourceResponse.FormMetadata,
        ],
    ],
] = {
    "process": ZeebeProcessAdapter._create_process_from_raw_process,
    "decision": ZeebeProcessAdapter._create_decision_from_raw_decision,
    "decisionRequirements": ZeebeProcessAdapter._create_decision_requirements_from_raw_decision_requirements,
    "form": ZeebeProcessAdapter._create_form_from_raw_form,
}


async def _create_process_request(process_file_path: str) -> ProcessRequestObject:
    async with aiofiles.open(process_file_path, "rb") as file:
        return ProcessRequestObject(name=os.path.basename(process_file_path), definition=await file.read())


async def _create_resource_request(resource_file_path: str) -> Resource:
    async with aiofiles.open(resource_file_path, "rb") as file:
        return Resource(name=os.path.basename(resource_file_path), content=await file.read())



import asyncio
import logging
from typing import List, Optional

from pyzeebe.errors import (
    ActivateJobsRequestInvalidError,
    ZeebeBackPressureError,
    ZeebeGatewayUnavailableError,
    ZeebeInternalError,
)
from pyzeebe.grpc_internals.zeebe_job_adapter import ZeebeJobAdapter
from pyzeebe.job.job import Job
from pyzeebe.task.task import Task
from pyzeebe.worker.task_state import TaskState

logger = logging.getLogger(__name__)


class JobPoller:
    def __init__(
        self,
        zeebe_adapter: ZeebeJobAdapter,
        task: Task,
        queue: "asyncio.Queue[Job]",
        worker_name: str,
        request_timeout: int,
        task_state: TaskState,
        poll_retry_delay: int,
        tenant_ids: Optional[List[str]],
    ) -> None:
        self.zeebe_adapter = zeebe_adapter
        self.task = task
        self.queue = queue
        self.worker_name = worker_name
        self.request_timeout = request_timeout
        self.task_state = task_state
        self.poll_retry_delay = poll_retry_delay
        self.tenant_ids = tenant_ids
        self.stop_event = asyncio.Event()

    async def poll(self) -> None:
        while self.should_poll():
            await self.activate_max_jobs()

    async def activate_max_jobs(self) -> None:
        if self.calculate_max_jobs_to_activate() > 0:
            await self.poll_once()
        else:
            logger.warning(
                "Maximum number of jobs running for %s. Polling again in %s seconds...",
                self.task.type,
                self.poll_retry_delay,
            )
            await asyncio.sleep(self.poll_retry_delay)

    async def poll_once(self) -> None:
        try:
            jobs = self.zeebe_adapter.activate_jobs(
                task_type=self.task.type,
                worker=self.worker_name,
                timeout=self.task.config.timeout_ms,
                max_jobs_to_activate=self.calculate_max_jobs_to_activate(),
                variables_to_fetch=self.task.config.variables_to_fetch or [],
                request_timeout=self.request_timeout,
                tenant_ids=self.tenant_ids,
            )
            async for job in jobs:
                self.task_state.add(job)
                await self.queue.put(job)
        except ActivateJobsRequestInvalidError:
            logger.warning("Activate job requests was invalid for task %s", self.task.type)
            raise
        except (ZeebeBackPressureError, ZeebeGatewayUnavailableError, ZeebeInternalError) as error:
            logger.warning(
                "Failed to activate jobs from the gateway. Exception: %s. Retrying in 5 seconds...",
                repr(error),
            )
            await asyncio.sleep(5)

    def should_poll(self) -> bool:
        return not self.stop_event.is_set() and (self.zeebe_adapter.connected or self.zeebe_adapter.retrying_connection)

    def calculate_max_jobs_to_activate(self) -> int:
        worker_max_jobs = self.task.config.max_running_jobs - self.task_state.count_active()
        return min(worker_max_jobs, self.task.config.max_jobs_to_activate)

    async def stop(self) -> None:
        self.stop_event.set()
        await self.queue.join()



from __future__ import annotations

import inspect
from typing import Any, List, Optional

from pyzeebe.function_tools import Function
from pyzeebe.job.job import Job


def get_parameters_from_function(task_function: Function[..., Any]) -> Optional[List[str]]:
    function_signature = inspect.signature(task_function)
    for _, parameter in function_signature.parameters.items():
        if parameter.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
            return []

    if not function_signature.parameters:
        return None

    if all(param.annotation == Job for param in function_signature.parameters.values()):
        return []

    return [param.name for param in function_signature.parameters.values() if param.annotation != Job]


def get_job_parameter_name(function: Function[..., Any]) -> Optional[str]:
    function_signature = inspect.signature(function)
    params = list(function_signature.parameters.values())
    for param in params:
        if param.annotation == Job:
            return param.name
    return None



import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Optional

from pyzeebe.errors import NoZeebeAdapterError
from pyzeebe.job.job_status import JobStatus
from pyzeebe.types import Variables

if TYPE_CHECKING:
    from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter


@dataclass
class Job:
    key: int
    type: str
    process_instance_key: int
    bpmn_process_id: str
    process_definition_version: int
    process_definition_key: int
    element_id: str
    element_instance_key: int
    custom_headers: Dict[str, Any]
    worker: str
    retries: int
    deadline: int
    variables: Variables
    tenant_id: Optional[str] = None
    status: JobStatus = JobStatus.Running
    zeebe_adapter: Optional["ZeebeAdapter"] = None

    async def set_running_after_decorators_status(self) -> None:
        """
        RunningAfterDecorators status means that the task has been completed as intended and the after decorators will now run.

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            self.status = JobStatus.RunningAfterDecorators
        else:
            raise NoZeebeAdapterError()

    async def set_success_status(self) -> None:
        """
        Success status means that the job has been completed as intended.

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            self.status = JobStatus.Completed
            await self.zeebe_adapter.complete_job(job_key=self.key, variables=self.variables)
        else:
            raise NoZeebeAdapterError()

    async def set_failure_status(
        self,
        message: str,
        retry_back_off_ms: int = 0,
        variables: Optional[Variables] = None,
    ) -> None:
        """
        Failure status means a technical error has occurred. If retried the job may succeed.
        For example: connection to DB lost

        Args:
            message (str): The failure message that Zeebe will receive
            retry_back_off_ms (int): The backoff timeout (in ms) for the next retry. New in Zeebe 8.1.
            variables (dict): A dictionary containing variables that will instantiate the variables at
                the local scope of the job's associated task. Must be JSONable. New in Zeebe 8.2.

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            self.status = JobStatus.Failed
            await self.zeebe_adapter.fail_job(
                job_key=self.key,
                retries=self.retries - 1,
                message=message,
                retry_back_off_ms=retry_back_off_ms,
                variables=variables or {},
            )
        else:
            raise NoZeebeAdapterError()

    async def set_error_status(
        self,
        message: str,
        error_code: str = "",
        variables: Optional[Variables] = None,
    ) -> None:
        """
        Error status means that the job could not be completed because of a business error and won't ever be able to be completed.
        For example: a required parameter was not given
        An error code can be added to handle the error in the Zeebe process

        Args:
            message (str): The error message
            error_code (str): The error code that Zeebe will receive
            variables (dict): A dictionary containing variables that will instantiate the variables at
                the local scope of the job's associated task. Must be JSONable. New in Zeebe 8.2.

        Raises:
            NoZeebeAdapterError: If the job does not have a configured ZeebeAdapter
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error

        """
        if self.zeebe_adapter:
            self.status = JobStatus.ErrorThrown
            await self.zeebe_adapter.throw_error(
                job_key=self.key, message=message, error_code=error_code, variables=variables or {}
            )
        else:
            raise NoZeebeAdapterError()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Job):
            raise NotImplementedError()
        return self.key == other.key


def create_copy(job: Job) -> Job:
    return Job(
        job.key,
        job.type,
        job.process_instance_key,
        job.bpmn_process_id,
        job.process_definition_version,
        job.process_definition_key,
        job.element_id,
        job.element_instance_key,
        copy.deepcopy(job.custom_headers),
        job.worker,
        job.retries,
        job.deadline,
        copy.deepcopy(job.variables),
        job.tenant_id,
        job.status,
        job.zeebe_adapter,
    )



import asyncio
from typing import List, Optional

import grpc
from typing_extensions import deprecated

from pyzeebe import ZeebeClient
from pyzeebe.grpc_internals.types import (
    CancelProcessInstanceResponse,
    CreateProcessInstanceResponse,
    CreateProcessInstanceWithResultResponse,
    DeployProcessResponse,
    DeployResourceResponse,
    PublishMessageResponse,
)
from pyzeebe.types import Variables


class SyncZeebeClient:
    def __init__(self, grpc_channel: grpc.aio.Channel, max_connection_retries: int = 10) -> None:
        self.loop = asyncio.get_event_loop()
        self.client = ZeebeClient(grpc_channel, max_connection_retries)

    def run_process(
        self,
        bpmn_process_id: str,
        variables: Optional[Variables] = None,
        version: int = -1,
        tenant_id: Optional[str] = None,
    ) -> CreateProcessInstanceResponse:
        return self.loop.run_until_complete(self.client.run_process(bpmn_process_id, variables, version, tenant_id))

    def run_process_with_result(
        self,
        bpmn_process_id: str,
        variables: Optional[Variables] = None,
        version: int = -1,
        timeout: int = 0,
        variables_to_fetch: Optional[List[str]] = None,
        tenant_id: Optional[str] = None,
    ) -> CreateProcessInstanceWithResultResponse:
        return self.loop.run_until_complete(
            self.client.run_process_with_result(
                bpmn_process_id, variables, version, timeout, variables_to_fetch, tenant_id
            )
        )

    def cancel_process_instance(self, process_instance_key: int) -> CancelProcessInstanceResponse:
        return self.loop.run_until_complete(self.client.cancel_process_instance(process_instance_key))

    @deprecated("Deprecated since Zeebe 8.0. Use deploy_resource instead")
    def deploy_process(self, *process_file_path: str) -> DeployProcessResponse:
        return self.loop.run_until_complete(self.client.deploy_process(*process_file_path))

    def deploy_resource(self, *resource_file_path: str, tenant_id: Optional[str] = None) -> DeployResourceResponse:
        return self.loop.run_until_complete(self.client.deploy_resource(*resource_file_path, tenant_id=tenant_id))

    def publish_message(
        self,
        name: str,
        correlation_key: str,
        variables: Optional[Variables] = None,
        time_to_live_in_milliseconds: int = 60000,
        message_id: Optional[str] = None,
        tenant_id: Optional[str] = None,
    ) -> PublishMessageResponse:
        return self.loop.run_until_complete(
            self.client.publish_message(
                name,
                correlation_key,
                variables,
                time_to_live_in_milliseconds,
                message_id,
                tenant_id,
            )
        )



import logging
from typing import List

from pyzeebe import Job

logger = logging.getLogger(__name__)


class TaskState:
    def __init__(self) -> None:
        self._active_jobs: List[int] = []

    def remove(self, job: Job) -> None:
        try:
            self._active_jobs.remove(job.key)
        except ValueError:
            logger.warning("Could not find Job key %s when trying to remove from TaskState", job.key)

    def add(self, job: Job) -> None:
        self._active_jobs.append(job.key)

    def count_active(self) -> int:
        return len(self._active_jobs)



from __future__ import annotations

import asyncio
import functools
from typing import Any, Iterable, List, TypeVar

from typing_extensions import ParamSpec, TypeIs

from pyzeebe.function_tools import AsyncFunction, Function, SyncFunction

P = ParamSpec("P")
R = TypeVar("R")


def asyncify_all_functions(functions: Iterable[Function[..., Any]]) -> List[AsyncFunction[..., Any]]:
    async_functions: list[AsyncFunction[..., Any]] = []
    for function in functions:
        if not is_async_function(function):
            async_functions.append(asyncify(function))
        else:
            async_functions.append(function)
    return async_functions


def asyncify(task_function: SyncFunction[P, R]) -> AsyncFunction[P, R]:
    @functools.wraps(task_function)
    async def async_function(*args: P.args, **kwargs: P.kwargs) -> R:
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, functools.partial(task_function, *args, **kwargs))

    return async_function


def is_async_function(function: Function[P, R]) -> TypeIs[AsyncFunction[P, R]]:
    # Not using inspect.iscoroutinefunction here because it doens't handle AsyncMock well
    # See: https://bugs.python.org/issue40573
    return asyncio.iscoroutinefunction(function)



__version__ = "4.0.0rc3"

from pyzeebe import errors
from pyzeebe.channel import *
from pyzeebe.client.client import ZeebeClient
from pyzeebe.client.sync_client import SyncZeebeClient
from pyzeebe.credentials.base import CredentialsABC
from pyzeebe.credentials.camunda_identity import CamundaIdentityCredentials
from pyzeebe.credentials.plugins import AuthMetadataPlugin
from pyzeebe.job.job import Job
from pyzeebe.job.job_status import JobStatus
from pyzeebe.task.exception_handler import ExceptionHandler, default_exception_handler
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import TaskDecorator
from pyzeebe.worker.task_router import ZeebeTaskRouter
from pyzeebe.worker.worker import ZeebeWorker

__all__ = (
    "errors",
    "create_camunda_cloud_channel",
    "create_insecure_channel",
    "create_secure_channel",
    "ZeebeClient",
    "SyncZeebeClient",
    "Job",
    "JobStatus",
    "ExceptionHandler",
    "TaskConfig",
    "TaskDecorator",
    "ZeebeTaskRouter",
    "default_exception_handler",
    "ZeebeWorker",
    "CredentialsABC",
    "CamundaIdentityCredentials",
    "AuthMetadataPlugin",
)



import logging
from typing import Awaitable, Callable

from pyzeebe.errors.pyzeebe_errors import BusinessError
from pyzeebe.job.job import Job

logger = logging.getLogger(__name__)

ExceptionHandler = Callable[[Exception, Job], Awaitable[None]]


async def default_exception_handler(e: Exception, job: Job) -> None:
    logger.warning("Task type: %s - failed job %s. Error: %s.", job.type, job, e)
    if isinstance(e, BusinessError):
        await job.set_error_status(f"Failed job. Recoverable error: {e}", error_code=e.error_code)
    else:
        await job.set_failure_status(f"Failed job. Error: {e}")



from __future__ import annotations

import functools
import logging
from typing import Any, Dict, Sequence, Tuple, TypeVar

from typing_extensions import ParamSpec

from pyzeebe import Job
from pyzeebe.function_tools import DictFunction, Function
from pyzeebe.function_tools.async_tools import asyncify, is_async_function
from pyzeebe.function_tools.dict_tools import convert_to_dict_function
from pyzeebe.function_tools.parameter_tools import get_job_parameter_name
from pyzeebe.job.job import create_copy
from pyzeebe.task.exception_handler import default_exception_handler
from pyzeebe.task.task import Task
from pyzeebe.task.task_config import TaskConfig
from pyzeebe.task.types import AsyncTaskDecorator, DecoratorRunner, JobHandler

P = ParamSpec("P")
R = TypeVar("R")

logger = logging.getLogger(__name__)


def build_task(task_function: Function[..., Any], task_config: TaskConfig) -> Task:
    task_config.job_parameter_name = get_job_parameter_name(task_function)
    return Task(task_function, build_job_handler(task_function, task_config), task_config)


def build_job_handler(task_function: Function[..., Any], task_config: TaskConfig) -> JobHandler:
    prepared_task_function = prepare_task_function(task_function, task_config)

    before_decorator_runner = create_decorator_runner(task_config.before)
    after_decorator_runner = create_decorator_runner(task_config.after)

    @functools.wraps(task_function)
    async def job_handler(job: Job) -> Job:
        if task_config.job_parameter_name:
            job.variables[task_config.job_parameter_name] = create_copy(job)

        job = await before_decorator_runner(job)
        original_return_value, succeeded = await run_original_task_function(prepared_task_function, task_config, job)
        job.variables.update(original_return_value)
        job.variables.pop(task_config.job_parameter_name, None)  # type: ignore[arg-type]
        await job.set_running_after_decorators_status()
        job = await after_decorator_runner(job)
        if succeeded:
            await job.set_success_status()
        return job

    return job_handler


def prepare_task_function(task_function: Function[P, R], task_config: TaskConfig) -> DictFunction[P]:
    if not is_async_function(task_function):
        task_function = asyncify(task_function)

    if task_config.single_value:
        return convert_to_dict_function(task_function, task_config.variable_name)
    # we check return type in task decorator
    return task_function  # type: ignore[return-value]


async def run_original_task_function(
    task_function: DictFunction[...], task_config: TaskConfig, job: Job
) -> Tuple[Dict[str, Any], bool]:
    try:
        if task_config.variables_to_fetch is None:
            variables = {}
        else:
            variables = {
                k: v
                for k, v in job.variables.items()
                if k in task_config.variables_to_fetch or k == task_config.job_parameter_name
            }
        returned_value = await task_function(**variables)

        if returned_value is None:
            returned_value = {}

        return returned_value, True
    except Exception as e:
        logger.debug("Failed job: %s. Error: %s.", job, e)
        exception_handler = task_config.exception_handler or default_exception_handler
        await exception_handler(e, job)
        return job.variables, False


def create_decorator_runner(decorators: Sequence[AsyncTaskDecorator]) -> DecoratorRunner:
    async def decorator_runner(job: Job) -> Job:
        for decorator in decorators:
            job = await run_decorator(decorator, job)
        return job

    return decorator_runner


async def run_decorator(decorator: AsyncTaskDecorator, job: Job) -> Job:
    try:
        return await decorator(job)
    except Exception as e:
        logger.warning("Failed to run decorator %s. Exception: %s", decorator, e, exc_info=True)
        return job



import grpc

from pyzeebe.credentials.base import CredentialsABC


class AuthMetadataPlugin(grpc.AuthMetadataPlugin):  # type: ignore[misc]
    """Custom authentication plugin with exception catching.

    Args:
        credentials (CredentialsABC): A credentials manager.
    """

    def __init__(self, *, credentials: CredentialsABC) -> None:
        self._credentials = credentials

    def __call__(self, context: grpc.AuthMetadataContext, callback: grpc.AuthMetadataPluginCallback) -> None:
        """Implements authentication by passing metadata to a callback.

        This method will be invoked asynchronously in a separate thread.

        Args:
          context: An AuthMetadataContext providing information on the RPC that
            the plugin is being called to authenticate.
          callback: An AuthMetadataPluginCallback to be invoked either
            synchronously or asynchronously.
        """
        try:
            metadata = self._credentials.get_auth_metadata(context)
        except Exception as e:
            callback((), e)
        else:
            callback(metadata, None)



import datetime
import threading
from typing import Any, Dict, Optional

import requests

from pyzeebe.credentials.base import CredentialsABC
from pyzeebe.credentials.typing import AuthMetadata, CallContext
from pyzeebe.errors import InvalidOAuthCredentialsError


class CamundaIdentityCredentials(CredentialsABC):
    """Credentials client for Camunda Platform.

    Args:
        oauth_url (str): The Keycloak auth endpoint url.
        client_id (str): The client id provided by Camunda Platform
        client_secret (str): The client secret provided by Camunda Platform
        audience (str): Audience for Zeebe. Default: zeebe-api
        refresh_threshold_seconds (int): Will try to refresh token if it expires in this number of seconds or less. Default: 20
    """

    def __init__(
        self,
        *,
        oauth_url: str,
        client_id: str,
        client_secret: str,
        audience: str = "zeebe-api",
        refresh_threshold_seconds: int = 20
    ) -> None:
        self.oauth_url = oauth_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.audience = audience

        self._lock = threading.Lock()
        self._refresh_threshold = datetime.timedelta(seconds=refresh_threshold_seconds)

        self._token: Optional[Dict[str, Any]] = None
        self._expires_in: Optional[datetime.datetime] = None

    def _expired(self) -> bool:
        return (
            self._token is None
            or self._expires_in is None
            or (self._expires_in - self._refresh_threshold) < datetime.datetime.now(datetime.timezone.utc)
        )

    def _refresh(self) -> None:
        try:
            response = requests.post(
                self.oauth_url,
                data={
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "audience": self.audience,
                    "grant_type": "client_credentials",
                },
            )
            response.raise_for_status()
            data = response.json()
            self._token = data["access_token"]
            self._expires_in = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
                seconds=int(data["expires_in"])
            )
        except requests.HTTPError as http_error:
            raise InvalidOAuthCredentialsError(
                url=self.oauth_url, client_id=self.client_id, audience=self.audience
            ) from http_error

    def get_auth_metadata(self, context: CallContext) -> AuthMetadata:
        """
        Args:
            context (grpc.AuthMetadataContext): Provides information to call credentials metadata plugins.

        Returns:
            Tuple[Tuple[str, Union[str, bytes]], ...]: The `metadata` used to construct the :py:class:`grpc.CallCredentials`.

        Raises:
            InvalidOAuthCredentialsError: One of the provided camunda credentials is not correct
        """
        with self._lock:
            if self._expired() is True:
                self._refresh()
            return (("authorization", "Bearer {}".format(self._token)),)



import asyncio
import logging
import socket
from typing import List, Optional

import grpc

from pyzeebe import TaskDecorator
from pyzeebe.grpc_internals.zeebe_adapter import ZeebeAdapter
from pyzeebe.job.job import Job
from pyzeebe.task import task_builder
from pyzeebe.task.exception_handler import ExceptionHandler
from pyzeebe.worker.job_executor import JobExecutor
from pyzeebe.worker.job_poller import JobPoller
from pyzeebe.worker.task_router import ZeebeTaskRouter
from pyzeebe.worker.task_state import TaskState

logger = logging.getLogger(__name__)


class ZeebeWorker(ZeebeTaskRouter):
    """A zeebe worker that can connect to a zeebe instance and perform tasks."""

    def __init__(
        self,
        grpc_channel: grpc.aio.Channel,
        name: Optional[str] = None,
        request_timeout: int = 0,
        before: Optional[List[TaskDecorator]] = None,
        after: Optional[List[TaskDecorator]] = None,
        max_connection_retries: int = 10,
        watcher_max_errors_factor: int = 3,
        poll_retry_delay: int = 5,
        tenant_ids: Optional[List[str]] = None,
        exception_handler: Optional[ExceptionHandler] = None,
    ):
        """
        Args:
            grpc_channel (grpc.aio.Channel): GRPC Channel connected to a Zeebe gateway
            name (str): Name of zeebe worker
            request_timeout (int): Longpolling timeout for getting tasks from zeebe. If 0 default value is used
            before (List[TaskDecorator]): Decorators to be performed before each task
            after (List[TaskDecorator]): Decorators to be performed after each task
            exception_handler (ExceptionHandler): Handler that will be called when a job fails.
            max_connection_retries (int): Amount of connection retries before worker gives up on connecting to zeebe. To setup with infinite retries use -1
            watcher_max_errors_factor (int): Number of consecutive errors for a task watcher will accept before raising MaxConsecutiveTaskThreadError
            poll_retry_delay (int): The number of seconds to wait before attempting to poll again when reaching max amount of running jobs
            tenant_ids (List[str]): A list of tenant IDs for which to activate jobs. New in Zeebe 8.3.
        """
        super().__init__(before, after, exception_handler)
        self.zeebe_adapter = ZeebeAdapter(grpc_channel, max_connection_retries)
        self.name = name or socket.gethostname()
        self.request_timeout = request_timeout
        self.watcher_max_errors_factor = watcher_max_errors_factor
        self._watcher_thread = None
        self.poll_retry_delay = poll_retry_delay
        self.tenant_ids = tenant_ids
        self._work_task: "Optional[asyncio.Future[List[None]]]" = None
        self._job_pollers: List[JobPoller] = []
        self._job_executors: List[JobExecutor] = []

    async def work(self) -> None:
        """
        Start the worker. The worker will poll zeebe for jobs of each task in a different thread.

        Raises:
            ActivateJobsRequestInvalidError: If one of the worker's task has invalid types
            ZeebeBackPressureError: If Zeebe is currently in back pressure (too many requests)
            ZeebeGatewayUnavailableError: If the Zeebe gateway is unavailable
            ZeebeInternalError: If Zeebe experiences an internal error
            UnknownGrpcStatusCodeError: If Zeebe returns an unexpected status code

        """
        self._job_executors, self._job_pollers = [], []

        for task in self.tasks:
            jobs_queue: "asyncio.Queue[Job]" = asyncio.Queue()
            task_state = TaskState()

            poller = JobPoller(
                self.zeebe_adapter,
                task,
                jobs_queue,
                self.name,
                self.request_timeout,
                task_state,
                self.poll_retry_delay,
                self.tenant_ids,
            )
            executor = JobExecutor(task, jobs_queue, task_state)
            self._job_pollers.append(poller)
            self._job_executors.append(executor)

        coroutines = [poller.poll() for poller in self._job_pollers] + [
            executor.execute() for executor in self._job_executors
        ]

        self._work_task = asyncio.gather(*coroutines)

        try:
            await self._work_task
        except asyncio.CancelledError:
            logger.info("Zeebe worker was stopped")
            return

    async def stop(self) -> None:
        """
        Stop the worker. This will emit a signal asking tasks to complete the current task and stop polling for new.
        """
        if self._work_task is not None:
            self._work_task.cancel()

        for poller in self._job_pollers:
            await poller.stop()

        for executor in self._job_executors:
            await executor.stop()

    def include_router(self, *routers: ZeebeTaskRouter) -> None:
        """
        Adds all router's tasks to the worker.

        Raises:
            DuplicateTaskTypeError: If a task from the router already exists in the worker

        """
        for router in routers:
            for task in router.tasks:
                config_with_decorators = self._add_decorators_to_config(task.config)
                task = task_builder.build_task(task.original_function, config_with_decorators)
                self._add_task(task)



from dataclasses import dataclass
from typing import List, Optional, Union

from pyzeebe.types import Variables


@dataclass(frozen=True)
class CreateProcessInstanceResponse:
    #: the key of the process definition which was used to create the process instance
    process_definition_key: int
    #: the BPMN process ID of the process definition which was used to create the process
    #: instance
    bpmn_process_id: str
    #: the version of the process definition which was used to create the process instance
    version: int
    #: the unique identifier of the created process instance; to be used wherever a request
    #: needs a process instance key (e.g. CancelProcessInstanceRequest)
    process_instance_key: int
    #: the tenant ID of the created process instance
    tenant_id: Optional[str]


@dataclass(frozen=True)
class CreateProcessInstanceWithResultResponse:
    #: the key of the process definition which was used to create the process instance
    process_definition_key: int
    #: the BPMN process ID of the process definition which was used to create the process
    #: instance
    bpmn_process_id: str
    #: the version of the process definition which was used to create the process instance
    version: int
    #: the unique identifier of the created process instance; to be used wherever a request
    #: needs a process instance key (e.g. CancelProcessInstanceRequest)
    process_instance_key: int
    #: consisting of all visible variables to the root scope
    variables: Variables
    #: the tenant ID of the process definition
    tenant_id: Optional[str]


@dataclass(frozen=True)
class CancelProcessInstanceResponse:
    pass


@dataclass(frozen=True)
class DeployProcessResponse:
    @dataclass(frozen=True)
    class ProcessMetadata:
        #: the bpmn process ID, as parsed during deployment; together with the version forms a
        #: unique identifier for a specific process definition
        bpmn_process_id: str
        #: the assigned process version
        version: int
        #: the assigned key, which acts as a unique identifier for this process
        process_definition_key: int
        #: the resource name (see: ProcessRequestObject.name) from which this process was
        #: parsed
        resource_name: str

    #: the unique key identifying the deployment
    key: int
    #: a list of deployed processes
    processes: List[ProcessMetadata]


@dataclass(frozen=True)
class DeployResourceResponse:
    @dataclass(frozen=True)
    class ProcessMetadata:
        #: the bpmn process ID, as parsed during deployment; together with the version forms a
        #: unique identifier for a specific process definition
        bpmn_process_id: str
        #: the assigned process version
        version: int
        #: the assigned key, which acts as a unique identifier for this process
        process_definition_key: int
        #: the resource name (see: ProcessRequestObject.name) from which this process was
        #: parsed
        resource_name: str
        #: the tenant ID of the deployed process
        tenant_id: Optional[str]

    @dataclass(frozen=True)
    class DecisionMetadata:
        #: the dmn decision ID, as parsed during deployment; together with the
        #: versions forms a unique identifier for a specific decision
        dmn_decision_id: str
        #: the dmn name of the decision, as parsed during deployment
        dmn_decision_name: str
        #: the assigned decision version
        version: int
        #: the assigned decision key, which acts as a unique identifier for this
        #: decision
        decision_key: int
        #: the dmn ID of the decision requirements graph that this decision is part
        #: of, as parsed during deployment
        dmn_decision_requirements_id: str
        #: the assigned key of the decision requirements graph that this decision is
        #: part of
        decision_requirements_key: int
        #: the tenant ID of the deployed decision
        tenant_id: Optional[str]

    @dataclass(frozen=True)
    class DecisionRequirementsMetadata:
        #: the dmn decision requirements ID, as parsed during deployment; together
        #: with the versions forms a unique identifier for a specific decision
        dmn_decision_requirements_id: str
        #: the dmn name of the decision requirements, as parsed during deployment
        dmn_decision_requirements_name: str
        #: the assigned decision requirements version
        version: int
        #: the assigned decision requirements key, which acts as a unique identifier
        #: for this decision requirements
        decision_requirements_key: int
        #: the resource name (see: Resource.name) from which this decision
        #: requirements was parsed
        resource_name: str
        #: the tenant ID of the deployed decision requirements
        tenant_id: Optional[str]

    @dataclass(frozen=True)
    class FormMetadata:
        #: the form ID, as parsed during deployment; together with the
        #: versions forms a unique identifier for a specific form
        form_id: str
        #: the assigned form version
        version: int
        #: the assigned key, which acts as a unique identifier for this form
        form_key: int
        #: the resource name
        resource_name: str
        #: the tenant ID of the deployed form
        tenant_id: Optional[str]

    #: the unique key identifying the deployment
    key: int
    #: a list of deployed resources, e.g. processes
    deployments: List[Union[ProcessMetadata, DecisionMetadata, DecisionRequirementsMetadata, FormMetadata]]
    #: the tenant ID of the deployed resources
    tenant_id: Optional[str]


@dataclass(frozen=True)
class PublishMessageResponse:
    #: the unique ID of the message that was published
    key: int
    #: the tenant ID of the message
    tenant_id: Optional[str]


@dataclass(frozen=True)
class CompleteJobResponse:
    pass


@dataclass(frozen=True)
class FailJobResponse:
    pass


@dataclass(frozen=True)
class ThrowErrorResponse:
    pass



from __future__ import annotations

from typing import Any, Awaitable, Callable, Dict, Optional, TypeVar, Union

from typing_extensions import ParamSpec

Parameters = ParamSpec("Parameters")
ReturnType = TypeVar("ReturnType")

SyncFunction = Callable[Parameters, ReturnType]
AsyncFunction = Callable[Parameters, Awaitable[ReturnType]]
Function = Union[SyncFunction[Parameters, ReturnType], AsyncFunction[Parameters, ReturnType]]

DictFunction = Callable[Parameters, Awaitable[Optional[Dict[str, Any]]]]
